#include "owl\owlpch.h"
#pragma hdrstop
#include <assert.h>

#include "lrstimer.h"

//////////////////////////////////////////////////////////////////
// Class LoResTimer

///////////////////////////
// static var & fn's

BOOL LoResTimer::Initialized = FALSE;
LoResTimer * LoResTimer::AllTimers[LRST_MAXTIMERS];
short LoResTimer::NumTimers;  // how many timers are there?
unsigned long LoResTimer::FastestTimerInterval;  // interval of fastest one, the one we use.
BOOL LoResTimer::DoingTick; // used for a few things.
BOOL LoResTimer::TimersChanged;
UINT LoResTimer::WinTimer;  // the timer we get from WIndows

BOOL LoResTimer::TimerSentry = 0;

//////////////////////////////////
// InitializeStatic()

void LoResTimer::InitializeStatic()
{
	if(Initialized)
		return;
	for(int x = 0; x < LRST_MAXTIMERS; x++)
		AllTimers[x] = NULL;
	NumTimers = NULL;
	FastestTimerInterval = -1;   // will give warning - we are really setting to MAXINT
	Initialized = TRUE;
	DoingTick = FALSE;
	WinTimer = NULL;
}

//////////////////////////////////
// TimerProcess()

// this has a dummy fn that calls it:
void FAR PASCAL MyTimerProc(HWND, UINT, UINT, DWORD)
{
	LoResTimer::TimerProcess();
}

// this is called by the Windows timer mechanism - it ticks all
// timers that need it.

// When in debugger, if you are single-stepping through this
// code and the timer fires again, it used to crash.  I guess the debugger
// doesn't like code executing when it thinks it has stopped the
// program.  So the sentry ignores any timer tick when its not done
// processing the previous one, allowing you to halt and debug a
// timer event.
void LoResTimer::TimerProcess()
{
	assert(Initialized);
	if(TimerSentry)
		return;
	TimerSentry = TRUE;
	if(!NumTimers)
	{
		TimerSentry = FALSE;
		return;
	}
	DoingTick = TRUE;  // so AddToStatic & RemoveFromStatic know whats up
	TimersChanged = FALSE;  // set by '''' if they changed the timers
	// loop thru all of em.
	unsigned long timenow = GetTickCount();
	for(int x = 0; x < LRST_MAXTIMERS; x++)
		if(AllTimers[x])
			AllTimers[x]->DoTick(timenow);

	// Now, if this process altered our timer list we may need to
	// modify our FastestTimer and our ticking.
	if(TimersChanged)
		UpdateFastestTimer();
	DoingTick = FALSE;

	TimerSentry = FALSE;
}

//////////////////////////////////
// ctors

// we ensure that the interval is never less than 55 msec.

LoResTimer::LoResTimer(TimerID id, unsigned long interval, HWND handle,
		MsgTimerStyle style)
	:	SmartTimer(id, interval, handle, style)
{
	// interval must be 55 <= interval <= 65535
	if(Interval < 55)
		Interval = 55;
	if(Interval > 65535l)
		Interval = 65535l;

	if(!Initialized)
		InitializeStatic();
}

LoResTimer::LoResTimer(TimerID id, unsigned long interval, SmartTimerFn fnptr)
	:	SmartTimer(id, interval, fnptr)
{
	// interval must be 55 <= interval <= 65535
	if(Interval < 55)
		Interval = 55;
	if(Interval > 65535l)
		Interval = 65535l;

	if(!Initialized)
		InitializeStatic();
}

//////////////////////////////////
// dtor

LoResTimer::~LoResTimer()
{
	if(Active)
		Stop();
}

//////////////////////////////////
// Start()

BOOL LoResTimer::Start()
{
	assert(Initialized);
	// this returns FALSE if any problems starting it up.
	return AddToStatic();
}

//////////////////////////////////
// Stop()

void LoResTimer::Stop()
{
	assert(Initialized);
	RemoveFromStatic();
}

//////////////////////////////////
// ResetInterval()

void LoResTimer::ResetInterval(unsigned long newinterval)
{
	assert(Initialized);
	assert(newinterval);
	if((Interval = newinterval) < 55)
		Interval = 55;
	if(Active)
	{
		// reset NextTick to value based on _now_
		NextTick = GetTickCount() + Interval;
		// if this is during processing of a tick, we let tick
		// processor worry about restarting timer, etc.
		if(DoingTick)
			TimersChanged = TRUE;
		else
			UpdateFastestTimer();
	}
}

//////////////////////////////////
// DoTick()

void LoResTimer::DoTick(unsigned long currenttime)
{
	assert(Initialized);
	// this is called often and must be fast.
	// Note that the user may delete us in the tick processing.
	// Hence, we must not refer to any member vars after calling it.


	if(NextTick > currenttime)
		return;
	// we must compensate for overrun - time past NextTick that
	// we missed. Add one interval, plus one for each one missed.
	// NextTick has to be in the future...
	unsigned long overrun = currenttime - NextTick;
	unsigned long missedintervals = overrun / Interval;
	NextTick += (1 + missedintervals) * Interval;
							 // ^^^^^^^^^^^^^^^^^^^^^^^^^^^ this truncates to long int
	switch(Style)
	{
		case FuncPtr:
			assert(FnPtr);
			if(FnPtr(ID))  // if it can continue
				return;
			else
				RemoveFromStatic();
			return;

		case SingleMsg:
			assert(Handle);
			if(SendMessage(Handle, ST_TimerMessage, (WPARAM) ID, NULL))  // if sent & can continue
				return;
			else
				RemoveFromStatic();
			return;

		case OffsetMsg:
			assert(Handle);
			if(SendMessage(Handle, ST_TimerMessage + ID, (WPARAM) ID, NULL))  // if sent & can continue
				return;
			else
				RemoveFromStatic();
			return;
	}
}

//////////////////////////////////
// AddToStatic()

BOOL LoResTimer::AddToStatic()
{
	assert(Initialized);
	// this finds a place for us and gets us going.
	if(NumTimers == LRST_MAXTIMERS)
		return FALSE;

	// Ok find a space for us.
	for(int x = 0; x < LRST_MAXTIMERS; x++)
		if(AllTimers[x] == NULL)  // if one there skip it
			break;

	if(x == LRST_MAXTIMERS)  // didnt find one?  ?
	{
		assert(0);  // for debug mode!
		NumTimers = LRST_MAXTIMERS;    // fix error otherwise
		return FALSE;
	}

	// ok we found it a place.  Put us there and set our NextTick.
	NextTick = GetTickCount() + Interval;
	AllTimers[x] = this;
	NumTimers++;
	if(DoingTick)
		TimersChanged = TRUE;
	else
		UpdateFastestTimer();

	Active = TRUE;
	return TRUE;
}


//////////////////////////////////
// RemoveFromStatic()

// we will attempt to do this without using member vars.

void LoResTimer::RemoveFromStatic()
{
	assert(Initialized);
	if(!NumTimers)
		return;

	// attempt to find it.
	for(int x = 0; x < LRST_MAXTIMERS; x++)
		if(AllTimers[x] == this)
			break;

	// may not find it - Not an error since this can be called many times.
	if(x == LRST_MAXTIMERS)
		return;

	// Ok we found us.
	AllTimers[x] = NULL;
	NumTimers--;
	if(DoingTick)
		TimersChanged = TRUE;
	else
		UpdateFastestTimer();
	Active = FALSE;
}

//////////////////////////////////
// UpdateFastestTimer()

void LoResTimer::UpdateFastestTimer()
{
	// this fn could be called in a number of cases:
	// 1: this is the first timer - we need to start a timer.
	// 2: The last timer has been removed - must kill timer.
	// 3: A timer has been added/removed which is not fastest.
	// 4: A timer has been added/removed which is fastest.
	// Note that 1 & 2 are really special cases of 4.

	// Lets check some special cases.
	if((NumTimers == 1) && (WinTimer == NULL))  // first one has been added!
	{
		// find that timer
		for(int x = 0; x < LRST_MAXTIMERS; x++)
			if(AllTimers[x])
				break;
		assert(x < LRST_MAXTIMERS);    // assert we found one
		// ok so x is the timer.  We need to set the windows timer to
		// choice us with x's timer speed.
		FastestTimerInterval = AllTimers[x]->GetInterval();
		WinTimer = SetTimer(NULL, NULL, FastestTimerInterval, MyTimerProc);
		assert(WinTimer);
		return;
	}
	if((NumTimers == 0) && (WinTimer))  // last one been deleted!
	{
		// we simply kill the timer.
		KillTimer(NULL, WinTimer);
		WinTimer = NULL;
		FastestTimerInterval = -1;  // will warn
		return;
	}
	assert(WinTimer);  // should never happen.
	// we may have to adjust the timer.  Search to see if one
	// is as fast or faster than the fastest one.
	unsigned long newfastest = -1;  // will warn - gets us MAXLONG
	unsigned short newone = LRST_MAXTIMERS;
	for(int x = 0; x < LRST_MAXTIMERS; x++)
		if(AllTimers[x])
			if(AllTimers[x]->GetInterval() < newfastest)
			{
				newfastest = AllTimers[x]->GetInterval();
				newone = x;
			}

	assert(newone < LRST_MAXTIMERS);  // shouldnt happen. There is at least one timer.

	// Ok now newone is fastest one, newfastest is fastest time.
	// 3 cases - its slower, faster, or equal to current timer.
	if(newfastest == FastestTimerInterval)  // easy one - equal
		return;  // no change!

	// in other 2 cases, we do same thing - kill old timer and
	// start new one.
	FastestTimerInterval = newfastest;
	KillTimer(NULL, WinTimer);
	WinTimer = NULL;
	WinTimer = SetTimer(NULL, NULL, FastestTimerInterval, MyTimerProc);
	assert(WinTimer);
}














