#ifndef LRSTIMER_H
#define LRSTIMER_H

#include "smttimer.h"


// this is the LoResTimer, based on SmartTimer.  It uses a single windows
// timer to keep time.  It has a static component that all LoResTimers share,
// this component keeps track of all timers and how long they have been waiting
// for a timer pulse.  This static component has _one_ timer, set to
// approximately the fastest LoResTimer in existence.

// Note that if the system is busy, and a tick or two is missed, they do
// not "pile up" - you will not receive a barrage of timer messages after a
// long pause.  This is to keep things sane.  Since these timer pileups
// happen when the system is loaded, there is no sense in loading it more.

// Since the granularity of the Windows timer is appx. 55 msec (18 ticks/sec),
// it is useless to request a shorter interval, or to expect much accuracy
// in that range.  The HiResTimer (coming later) will handle this better.

// Our static component has its single timer set to the fastest timer in
// its records, and on each tick it determines which of the timers merit
// a tick - this is done by "timestamping" each timer, and comparing the
// current time (GetTickCount()) with the time stamp.  Simply stamping
// with a "last sent" will not work - slower ones will lose out on ticks
// over time.  Instead, we stamp for a "next tick" time - the time we should
// get another tick.  We only increment this time by a multiple of the
// intended interval.  (multiple is one usually, if ticks were missed,
// it is number of ticks - in essence, NextTick is always in the future.)

// Static Component keeps all timers in an array of fixed length.  This
// imposes a cap on how many timers can be active, but will speed processing.
// Right now I will set maximum timers at 256 - but we can recompile to
// increase this if needed.  Timers are only added to records when
// they are started, and they are removed when they are stopped.
// This keeps the number down - only active timers are counted or
// accounted for.  The first timer to be created initializes the
// static component - this only needs to be done once.

const int LRST_MAXTIMERS = 256;

class LoResTimer : public SmartTimer
{
public:
	LoResTimer(TimerID id, unsigned long interval, HWND handle,
			MsgTimerStyle style = SingleMsg);

	LoResTimer(TimerID id, unsigned long interval, SmartTimerFn fnptr);

	virtual ~LoResTimer();

	// starts timer.  Adds to static array & system
	virtual BOOL Start();   // returns TRUE if started OK.
	// stops it.  Removes it from static system.
	virtual void Stop();

	virtual void ResetInterval(unsigned long newinterval);

	// timer proc can call protected & private stuff.
	friend void FAR PASCAL MyTimerProc(HWND, UINT, UINT, DWORD);

protected:
	// this is used for timing. - is set at Start and on ticks.
	unsigned long NextTick;

	// this is called by static component on a tick - Timer should determine
	// if it needs to tick, and should update its own NextTick field.
	// Shoudl Stop() itself if it needs to.
	void DoTick(unsigned long currenttime);

	BOOL AddToStatic();  // adds self - called by Start()
	void RemoveFromStatic();  // removes self - called by Stop()

	// Static stuff
	static BOOL Initialized;   // is TRUE if we have been initialized
	static LoResTimer * AllTimers[LRST_MAXTIMERS];
	static short NumTimers;  // how many timers are there?
	static unsigned long FastestTimerInterval;  // interval of fastest one, the one we use.
	static BOOL DoingTick; // used for a few things.
	static BOOL TimersChanged;
	static UINT WinTimer;  // the timer we get from WIndows
	static BOOL TimerSentry;

	static void InitializeStatic();  // called once to set up static stuff.
	static void TimerProcess(); // called by timer to tick us.
	static void UpdateFastestTimer();  // updates which is fastest and may change timer

};











#endif