#ifndef SMTTIMER_H
#define SMTTIMER_H

// this is the abstract base class, SmartTimer.  We will derive
// concrete classes LoResTimer and HiResTimer from it that
// use the Windows Timer, or the MultiMedia timer, respectively.
// This timer can either call a function via a pointer, or it can
// SendMessage() to a HWND.  Returning 0 to the fn, or SendMessage(),
// (or SendMessage beig sent to invalid HWND) will stop the timer.

// When you set up the timer, you specify an ID, which is a unsigned short.
// This can be used in a number of ways.  If you are using a fn pointer,
// this ID will be the single argument of the function when it is called.
// If you use SendMessage(), you have two choices - you can have all your
// timer messages come through the same message, with the WPARAM indicating
// the Timer ID, or you can have the TimerID be added to the message ID,
// which will enable you to have separate message response fn's, one
// for each ID.  I could even provide a EV_SMARTTIMER message response macro
// you can use with your prefered timer ID. (This is left as an exercise to you)
// Default message action is with the single message.

// the ID we use
typedef unsigned short TimerID;

// the base message sent by the timer.
const int ST_TimerMessage = WM_USER + 0x500;  // out of range of most constants

// prototype for the function pointer.  This fn should return FALSE to
// automatically stop the timer.  This looks weird!  Basically a
// BOOL function that takes a TimerID as its only parm, is defined as a
// SmartTimerFn.
typedef BOOL (*SmartTimerFn)(TimerID);

class SmartTimer
{
public:
	// an enum useful to us
	enum MsgTimerStyle {FuncPtr, SingleMsg, OffsetMsg};

	// creates a timer with Style == FuncPtr
	SmartTimer(TimerID id, unsigned long interval, SmartTimerFn fnptr);
	// creates timer with specified style: SingleMsg or OffsetMsg
	SmartTimer(TimerID id, unsigned long interval, HWND handle,
			MsgTimerStyle style = SingleMsg);

	// stops it if not already stopped.
	virtual ~SmartTimer();

	// starts timer.
	virtual BOOL Start()=0;   // returns TRUE if started OK.
	// stops it
	virtual void Stop()=0;

	// lets you change interval while it is going. Can be called before or after
	// timer has been started, and even during processing of a timer tick.
	// The timer will do its best to ensure that the next tick comes on that
	// interval.
	virtual void ResetInterval(unsigned long newinterval)=0;
	inline unsigned long GetInterval() const;  // returns interval
	inline TimerID GetID() const;

protected:
	TimerID ID;
	unsigned long Interval;
	MsgTimerStyle Style;
	HWND Handle;
	SmartTimerFn FnPtr;
	BOOL Active;   // is the timer ticking?
};















#endif